#include <sys/time.h>

#import "DUView.h"
#import "DUWindowController.h"

#include "fileops.h"

@implementation DUView

static double gettime() {
  struct timeval tv;
  gettimeofday(&tv,NULL);
  return tv.tv_sec*1e3+tv.tv_usec*1e-3;
}

- (id)initWithFrame:(NSRect)frame {
  [super initWithFrame:frame];
  
  x_scale=150;
  y_scale=1.0/1024;
 
  data_color=[[NSColor colorWithCalibratedRed:.7 green:.7 blue:1 alpha:1] retain];
  rsrc_color=[[NSColor colorWithCalibratedRed:.7 green:1 blue:.7 alpha:1] retain];
  padding_color=[[NSColor colorWithCalibratedRed:.7 green:.7 blue:.7 alpha:1] retain];
  
  [self registerForDraggedTypes: [NSArray arrayWithObject:NSFilenamesPboardType]];

  state=St_empty;
  return self;
}


- (IBAction)setYZoom:(id)sender {
  double fact=[sender doubleValue];
  y_scale=1.0/(exp(fact*log(1024)));
  [self handleSizeChange];
  [self setNeedsDisplay: YES];
}

- (IBAction)open:(id)sender {
  NSOpenPanel *oPanel = [NSOpenPanel openPanel];
  NSString *nsFileName;

  if(state==St_loading) return;

  [oPanel setCanChooseFiles:NO];
  [oPanel setCanChooseDirectories:YES];
  [oPanel setAllowsMultipleSelection:NO];
   
  if([oPanel runModalForDirectory:NSHomeDirectory() 
      file:nil 
      types:nil] == NSOKButton)  {
    NSArray *filesToOpen = [oPanel filenames];
    nsFileName = [filesToOpen objectAtIndex:0];
  } else return;
  
  if(![self loadDirectory:nsFileName]) {
    NSRunAlertPanel(
        @"Alert!", 
        @"Could not explore dir...",
        @"Ok",nil,nil,nil);
  }
}

- (BOOL) loadDirectory:(NSString*)nsFileName {
  [nsFileName 
      getFileSystemRepresentation:root_name
      maxLength:1024];
  // printf("root_node=%s\n",root_name);

  if(state==St_ready) {
    dealloc_tree(root_node);
  }
#if 0
  // monotasking
  root_node=explore_tree(root_name,NULL);
  if(!root_node) return NO;
  
  [[self window] setTitle:nsFileName];

  state=St_ready;
  [self setNeedsDisplay:YES];

#else
  // mutitasking
  spy=malloc(sizeof(struct ProgressSpy));
  root_node=explore_tree(root_name,spy);
  if(!root_node) {
    free(spy);
    return NO;
  }
  
  [[self window] setTitle:nsFileName];

  state=St_loading;

  progressTimer=[NSTimer scheduledTimerWithTimeInterval:.2
      target:self
      selector:@selector(handleTimer:)
      userInfo:nil
      repeats:TRUE];
      
  [progressWin makeKeyAndOrderFront:self];
#endif

  return YES;
}

- (IBAction)openInWindow:(id)sender {
  [self open:sender];
}

- (void)handleTimer:(id)sender {
  assert(state==St_loading);

  assert(pthread_mutex_lock(&spy->mutex)==0);

  if(spy->percent==101)
    state=St_ready;
  else {
    spy->want_snap=1;
    pthread_cond_wait(&spy->cond,&spy->mutex);
    // printf("percent=%g\n",spy->percent);
    [progressPercent setDoubleValue:spy->percent];
    [progressFileName setStringValue:
        [NSString stringWithUTF8String:spy->current_name]];  
  }

  assert(pthread_mutex_unlock(&spy->mutex)==0);
  
  if(state==St_ready) {
    
    free(spy);
    [progressTimer invalidate];
    
    [progressWin close];
      
    [self handleSizeChange];
    [self handleFontChange];
    // [[self window] useOptimizedDrawing:YES];
    [self setNeedsDisplay: YES];
  }
}

- (void) handleSizeChange {
  NSPoint pt_i=[self visibleRect].origin;

  if(state!=St_ready) return;
  
  [self setFrame:NSMakeRect(
      0,0,
      (root_node->depth+2.5)*x_scale,
      y_scale*root_node->size*512+10)];
  
  if(prev_y_scale!=0) {
    double o_i=(pt_i.y-5.0)/prev_y_scale;
    pt_i.y=5.0+o_i*y_scale;
    [self scrollPoint: pt_i];
  }  

  prev_y_scale=y_scale;
}

- (void) handleFontChange {

  font=[NSFont fontWithName:@"Monaco" size:10.0];

  font_attrs=[[NSDictionary alloc] initWithObjectsAndKeys:
      font, NSFontAttributeName,
      nil];
}

static char *unit_names[]={
  "byte","kB","MB","GB"
};

static char *nameOfSize(double size) {
  static char buf[1024];
  int n;
  for(n=0;n<3;n++) {
    if(size<1024) break;
    size/=1024;
  }
  sprintf(buf,"%.3g %s",size,unit_names[n]);
  return buf;
}


- (void)recursiveDraw:(struct FileNode*)fn
    base:(double) x_base :(double) y_base
    y_range: (double) y_min: (double) y_max {
  double height=fn->size*y_scale*512;
  double width=fn->flags & FN_STAT_DIR ?
    x_scale*.95 : x_scale*1.2;

  if(height==0 || y_base>y_max || y_base+height<y_min) 
    return;

  if(height<1) {
    // Column is too short : just draw a line

    [[NSColor blackColor] set];

    [NSBezierPath strokeRect:NSMakeRect(
        x_base,y_base,x_scale*(fn->depth+1),0)];
    
/*
    if(fn->depth>5) 
      printf("%s : depth>5\n ",fn->name);
*/

    /*
    [NSBezierPath strokeLineFromPoint:
            NSMakePoint(x_base,y_base)
        toPoint:
            NSMakePoint(x_base+x_scale*(fn->depth+1),y_base)];
    */
  } else {
    { // draw rectangle
      double y0=MAX(y_base,y_min-10);
      double y1=MIN(y_base+height,y_max+10);
      double y_dr=y_base+fn->d_size*y_scale;
      double y_rp=y_dr+fn->r_size*y_scale;
      
      if(y_dr>y0) {
        [data_color set];
        [NSBezierPath fillRect: NSMakeRect(
            x_base,y0,width,MIN(y1,y_dr)-y0)];      
      }
      
      if(y_rp>y_dr+1 && !(y_rp<y0 || y_dr>y1)) {
        [rsrc_color set];
        [NSBezierPath fillRect: NSMakeRect(
            x_base,y_dr,width,y_rp-y_dr)];
      }
      
      if(y_base+height>y_rp+1 && !(y_base+height<y0 || y_rp>y1)) {
        [padding_color set];
        [NSBezierPath fillRect: NSMakeRect(
            x_base,y_rp,width,y_base+height-y_rp)];
      }
    
      if(selectedDepth && selectedStack[selectedDepth-1]==fn) 
        [[NSColor redColor] set];
      else 
        [[NSColor blackColor] set];
      [NSBezierPath strokeRect: NSMakeRect(
          x_base,y0,
          width,y1-y0)];
    }
    
    if(height>5) { // draw text
      double y0=MAX(y_base,y_min);
      double y1=MIN(y_base+height,y_max);
      NSRect text_clip;
      NSString *name=[NSString stringWithUTF8String:
          fn->name];
      NSString *text=[NSString localizedStringWithFormat:
          @"%@\n%s",name,nameOfSize(fn->size*512.0)];

      text_clip=NSMakeRect(
          x_base,y0,
          width,y1-y0);
      [text drawInRect:text_clip withAttributes: font_attrs];
    }
    if((fn->flags & FN_STAT_DIR) && 
      !(fn->flags & FN_MOUNT_POINT) &&
      !(fn->flags & FN_OPENDIR_ERROR)) { 
      struct FileNode *ch;
      for(ch=fn->children;ch;ch=ch->next) {
        [self recursiveDraw:ch
            base :x_base+x_scale :y_base
            y_range :y_min :y_max];
        y_base+=ch->size*512*y_scale;    
      }
    }    
  }  
}



- (BOOL)isOpaque {
  return NO;
}

- (void)drawRect:(NSRect)rect {
  double t0,t1;
  /*
  [[NSColor whiteColor] set];
  NSRectFill(rect);
  */
  if(state!=St_ready) return;
  if([self inLiveResize]) return;  
  
/*  [[NSGraphicsContext currentContext] setShouldAntialias:NO];*/
  
  t0=gettime();
  [self recursiveDraw:root_node
      base:0.0 :5.0
      y_range :rect.origin.y
              :rect.origin.y+rect.size.height];
  t1=gettime();
//  printf("time=%5.1f ms\n",t1-t0); 

}

- (void)setNeedsDisplayInRect:(NSRect)invalidRect {
  // No partial redraws
  [super setNeedsDisplayInRect:[self visibleRect]];
}

- (void)viewDidEndLiveResize {
  [self setNeedsDisplay:YES];
}

// selection 


- (int)recursiveSelect:(struct FileNode*)fn
    depth: (int)depth
    point: (double)x_pt:(double)y_pt
    base:(double) x_base :(double) y_base {
  double height=fn->size*y_scale*512;
  double width=fn->flags & FN_STAT_DIR ?
    x_scale*.95 : x_scale*1.2;
  
  selectedStack[depth]=fn;
  
  if(x_pt<=x_base+width) 
    return depth+1;
    
  if(height==0) return 0;
  
  if((fn->flags & FN_STAT_DIR) && 
    !(fn->flags & FN_MOUNT_POINT) &&
    !(fn->flags & FN_OPENDIR_ERROR)) { 
    struct FileNode *ch;
    for(ch=fn->children;ch;ch=ch->next) {
      double ch_height=ch->size*512*y_scale;
      if(y_pt<y_base+ch_height) 
        return [self recursiveSelect:ch
          depth:depth+1
          point:x_pt:y_pt
          base :x_base+x_scale :y_base];
      y_base+=ch_height;
    }
  }
  // falled through: point is in empty space behind a dir
  return 0;
}


- (void) mouseDown:(NSEvent *)event {
  NSPoint pt = 
      [self convertPoint:[event locationInWindow]
          fromView:nil];  
  if(state!=St_ready) return;
  
  selectedDepth=[self 
      recursiveSelect:root_node
      depth:0
      point:pt.x:pt.y
      base:0.0 :5.0];
  if(selectedDepth) {
    char *p=selectedName,*r;
    int d;
    for(r=root_name;*r;r++) 
      *p++=*r;
    for(d=1;d<selectedDepth;d++) {
      *p++='/';
      for(r=selectedStack[d]->name;*r;r++) 
        *p++=*r;
    }
    *p=0;
//    printf("selectedName=%s\n",selectedName);
  }
  
  [self setNeedsDisplay:YES];
}


// make sure menu items are seen if there is a selection
- (BOOL)validateMenuItem:(id <NSMenuItem>)menuItem {
  if([menuItem tag]>=23 && [menuItem tag]<30) // directory menu
    return selectedDepth!=0;
  return YES;
}

- (IBAction)showSelectedDir:(id)sender {
  
}

- (IBAction)revealSelectedDir:(id)sender {
  [[NSWorkspace sharedWorkspace] openFile:[NSString stringWithUTF8String:selectedName] 
      withApplication:@"Finder"];
}

- (IBAction)refreshSelectedDir:(id)sender {

}

// Drop destination

- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender {
  if([[[sender draggingPasteboard] types] containsObject:NSFilenamesPboardType]) {
    [[self enclosingScrollView] setBorderType:NSLineBorder];
    return NSDragOperationLink;
  }
  return NSDragOperationNone;
}

- (void)draggingExited:(id <NSDraggingInfo>)sender {
  [[self enclosingScrollView] setBorderType:NSBezelBorder];
}

- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender {
  return [[[sender draggingPasteboard] types] containsObject:NSFilenamesPboardType];
} 

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender {
  [[self enclosingScrollView] setBorderType:NSBezelBorder];
  if([[[sender draggingPasteboard] types] containsObject:NSFilenamesPboardType]) {
    id p=[[sender draggingPasteboard] propertyListForType:NSFilenamesPboardType];
    NSEnumerator *enumerator = [(NSArray*)p objectEnumerator];
    id obj;
    while ((obj = [enumerator nextObject])) {
      if([obj isKindOfClass: [NSString class]]) 
        return [self loadDirectory: (NSString*)obj];
     }
  }
  return NO;
}

// drag away

- (void)mouseDragged:(NSEvent *)event {
  //get the Pasteboard used for drag and drop operations
  NSPasteboard* dragPasteboard=[NSPasteboard pasteboardWithName:NSDragPboard];
  NSString *text=[NSString stringWithUTF8String:selectedName];
  //create a new image for our semi-transparent drag image
  NSImage* dragImage=[[NSImage alloc] initWithSize:[text sizeWithAttributes:nil]]; 
  NSPoint atPoint;

  [dragPasteboard declareTypes:[NSArray arrayWithObject:NSFilenamesPboardType] owner:self];
  [dragImage lockFocus];
  [text drawAtPoint:NSMakePoint(0,0) withAttributes:nil];    
  [dragImage unlockFocus];

  atPoint=[self convertPoint:[event locationInWindow] fromView:nil];
  atPoint.x-=[dragImage size].width/2;
  atPoint.y-=[dragImage size].height/2;


  //execute the drag
  [self dragImage: dragImage//image to be displayed under the mouse
      at: atPoint//point to start drawing drag image
      offset: NSMakeSize(0,0)//no offset, drag starts at mousedown location
      event: event//mousedown event
      pasteboard:dragPasteboard//pasteboard to pass to receiver
      source: self//object where the image is coming from
      slideBack: YES];//if the drag fails slide the icon back
  [dragImage release];//done with our dragImage
}

- (NSDragOperation)draggingSourceOperationMaskForLocal:(BOOL)flag {
  return NSDragOperationGeneric;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)event {
  return YES;//so source doesn't have to be the active window
}

- (void)pasteboard:(NSPasteboard *)sender provideDataForType:(NSString *)type {
 if([type compare: NSFilenamesPboardType]==NSOrderedSame) {
   [sender 
      setPropertyList:[NSArray arrayWithObject:[NSString stringWithUTF8String:selectedName]]
      forType:NSFilenamesPboardType];
  }
}



@end
